// Crypt7KE.cpp

#include "stdafx.h"
#include <malloc.h>


#ifndef uint8
#define uint8  unsigned char
#endif

#ifndef uint32
#define uint32 unsigned long int
#endif

typedef struct
{
    uint32 total[2];
    uint32 state[8];
    uint8 buffer[64];
}
sha256_context;

void sha256_starts( sha256_context *ctx );
void sha256_update( sha256_context *ctx, uint8 *input, uint32 length );
void sha256_finish( sha256_context *ctx, uint8 digest[32] );


BYTE bKeyBinary[32];
int iKeyIx = 0;
static __int64 llLastKeyPart = 0;
static int iKeyCntr = 0;
static DWORD dwDivKey = 0;
static DWORD dwDivRef = 0;
static PSTR pszPasswordPtr = NULL;
static int iPasswordPtrLen = 0;
static __int64 llBlockCnt = 0;


#define GET_KEYBYTE(p)							\
{												\
	if (iKeyIx==0) NextBinKey();				\
	*(unsigned char*)(p) = bKeyBinary[iKeyIx];	\
	iKeyIx = (iKeyIx + 1) & 0x1f;				\
}

void NextBinKey()
{
	sha256_context sha_ctx;
	if (iKeyCntr <= 0)
	{
		((PDWORD)bKeyBinary)[0] = 0xB1599535;
		((PDWORD)bKeyBinary)[1] = 0x3B9BC555;
		((PDWORD)bKeyBinary)[2] = 0x93A8B0B4;
		((PDWORD)bKeyBinary)[3] = 0xAE173BF4;
		((PDWORD)bKeyBinary)[4] = 0x2549D565;
		((PDWORD)bKeyBinary)[5] = 0xC9BBABD3;
		((PDWORD)bKeyBinary)[6] = 0x61C81A27;
		((PDWORD)bKeyBinary)[7] = 0x57CA2D0F;
		iKeyCntr = 0;
	}
	sha256_starts(  &sha_ctx);
	sha256_update(  &sha_ctx, bKeyBinary, 32);
	sha256_update(  &sha_ctx, (unsigned char*)pszPasswordPtr, (unsigned long)iPasswordPtrLen);
	sha256_finish(  &sha_ctx, bKeyBinary);
	iKeyCntr++;
}

void InitBinKey(char* szPassword, int iPasswordLen = 0)
{
	pszPasswordPtr = szPassword;	// Passwort fr SHA256-Key-Generator NextBinKey()
	iPasswordPtrLen = iPasswordLen;
	if (iPasswordPtrLen <= 0) iPasswordPtrLen = strlen(pszPasswordPtr);
	iKeyCntr = 0;	// Inistialisiert NextBinKey() beim nchsten Aufruf
	iKeyIx = 0;		// Inistialisiert GET_KEYBYTE(p) beim nchsten Aufruf
	dwDivKey = 0;	// Key-Divisor fr CreatePermutTab() initialisieren
	dwDivRef = 0;	// Referenz-Divisor fr CreatePermutTab() initialisieren
}

void InjectBinKey3(DWORD dwData1, DWORD dwData2, DWORD dwData3)
{
	((PDWORD)bKeyBinary)[0] ^= dwData1;
	((PDWORD)bKeyBinary)[1] ^= dwData2;
	((PDWORD)bKeyBinary)[2] ^= dwData3;
	iKeyIx = 0;
}

struct MWC8Stat { DWORD a[8]; DWORD x[8]; DWORD c[8]; }
	tMWCStat = { 4164903690,4204114314,4187999619,4198054089,4210396968,4197302403,4178097609,4201298934 };	// Faktor a
// Sichere Primzahlen: a*b-1 und a*b/2-1 sind Primzahlen, mit a=tMWCStat.a[i] und b=2^32, Test z.B.: https://www.wolframalpha.com/input/?i=Is+(4164903690+*+2%5E31+-+1)+prime%3F
// Damit ist die Periode des MWC-Pseudozufallsgenerators = a*b/2-1  (MWC: Multiply With Carry: t=a*x+c, x=t%b, c=(int)(t/b)
// Durch Verschachtelung von 4 Generatoren mit leicht unterschiedlichen Perioden erhlt man eine Periode von ungefhr 4*(a*b/2-1)^4 = 2.6*10^76

__forceinline DWORD GetNextCryptModB(int iSeedIx, DWORD dwInject)
{
	// 2 Vierfach-MWC-Zufallsgeneratoren, Periode je 2.6e76
	// MWC: Multiply With Carry: t=a*x+c+i, x=t%b, c=(int)(t/b), b=1<<32, Return: x
	// iSeedIx: 0..3 fr Generator 1, 4..7 fr Generator 2
	// dwInject: Zustzlicher Modifikator, ein konstanter Wert ndert die Zufallsreihe, nicht aber die Periode
  #ifdef NO_ASM
	unsigned __int64 t;
	PDWORD pdwStat = tMWCStat.x + iSeedIx;
	t = ((__int64)pdwStat[-8] * *pdwStat) + pdwStat[8] + dwInject;	// t=a*x+c+i
	pdwStat[8] = (DWORD)(t >> 32);		// c=(int)(t/b)
	*pdwStat = (DWORD)t;				// x=t%b
	return *pdwStat;
  #else		// Dasselbe in Assembler, etwas schneller
	__asm
	{
		mov ecx,iSeedIx
		lea esi,[ecx*4 + tMWCStat.x]
		mov ebx,[esi]			; Status x
		mov eax,[esi - 32]		; Faktor a
		mul ebx					; a*x -> edx:eax
		add eax,[esi + 32]		; Carry c
		adc edx,0
		add eax,dwInject
		adc edx,0
		mov [esi + 32],edx		; Carry c
		mov [esi],eax
	}
  #endif
}


DWORD GetNextCryptModA(int iSeedIx, DWORD dwInject)
{
	return GetNextCryptModB(iSeedIx, dwInject);
}


void FillBinKey(PBYTE pyData, int iBytes)
{
	if (pyData==NULL) pyData = (PBYTE)tMWCStat.x, iBytes = 64;
	for (int i=0; i<iBytes; i++)
		GET_KEYBYTE(pyData + i);
}


int CreatePermutTab(PBYTE pbyTab, PBYTE pbyITab, int iSize)
{
	int i, le;
	le = iSize;
	if (iSize<0) iSize=-iSize;
	if (pbyITab) for (i=0; i<iSize; i++) pbyTab[i] = pbyITab[i] = i;
	else	     for (i=0; i<iSize; i++) pbyTab[i] = i;
	if (le<0) return 0;

	for (i=0; i<iSize-1; i++)
	{
		while (!(dwDivRef & 0xff000000))
		{
			dwDivRef = (dwDivRef<<8) | 0xff;
			dwDivKey <<= 8;
			GET_KEYBYTE(&dwDivKey);
		}
		DWORD dwKRem = dwDivKey % (iSize-i);
		dwDivKey /= (iSize-i);
		dwDivRef /= (iSize-i);
		BYTE byTmp = pbyTab[i];
		pbyTab[i] = pbyTab[dwKRem+i];
		pbyTab[dwKRem+i] = byTmp;
	}
	if (pbyITab) for (i=0; i<iSize; i++)
		pbyITab[pbyTab[i]] = i;
	return 0;
}


/*
7KE-Verschsselung (7 KBit Encryption):
Dieses Verschlsselungsverfahren verwendet beliebige eindeutige und umkehrbare Abbildungen (auch bijektiv oder eineindeutig)
zwischen zwei 8-Bit-Werten (bijektive Selbstabbildung oder Permutation - nicht zu verwechseln mit Transposition).
Dabei gibt es 256! (Fakultt von 256) mgliche Abbildungen, das sind fast 10507 oder 21684 Abbildungen
(zum Vergleich, nur 256 von all diesen Abbildungen beschreiben eine XOR-Operation).
Zur Erstellung einer Abbildungstabelle sind 1684 Bit erforderlich (GammaLog(1+256)/log(2)=1684).
Von all den 10507 mglichen Abbildungen werden vllig willkrlich vier herausgegriffen, und zwar abhngig vom Schlssel.
Dafr kommen 4 * 1684 = 6736 Bits vom insgesamt 7280 Bit groen Schlssel zum Einsatz.
Durch Modifikation und Verkettung entstehen daraus 266 = 7 * 1019 unterschiedliche Abbildungen, das reicht,
um selbst bei 64 Exabyte (64 Millionen Terrabyte) jedes Byte mit einer einzigartigen Abbildung
kodieren zu knnen, so dass statistische Analysen von vorn herein zum Scheitern verurteilt sind.

Welches Byte im Datenblock mit welcher Abbildung kodiert wird, ist abhngig vom Schlssel,
von den unverschlsselten und den verschlsselten Daten und zwei Pseudozufallsgeneratoren
mit jeweils 256 Bit groem "Seed" (Status je 256 Bit, Periode je etwa 2.6*1076).
Deren Initialisierungswerte (Seeds) sind ebenfalls Teil des Schlssels, so dass sich zusammen mit
32 Initialisierungsbits eine Schlsselgre von 4 * 1684 + 2 * 256 + 32 = 7280 Bit ergibt.
Diese werden aus einem beliebig groen Passwort durch wiederholte Anwendung der Hash-Operation
SHA-256 gewonnen, s. a. CalcSha256.

Die Blockgre ist beliebig, die Daten eines Blocks mssen aber komplett im Speicher stehen.
Damit die nderung nur eines Bits im Block ein vollkommen unterschiedliches Ergebnis zur Folge hat,
wird jeder Block vorwrts und rckwrts durchlaufen mit jeweils unterschiedlichen Parametern
(daher auch zwei Pseudozufallsgeneratoren, je einer fr vorwrts und rckwrts).
Sind die Daten zu gro, um sie komplett in dem Speicher zu laden, kann man sie in Blcke unterteilen.
Dann sorgt der Parameter dBlockNum dafr, dass jeder Block auf andere Weise verschlsselt wird.
Der Befehl CryptFile verwendet bei Dateien ab einer Gre von 4096 Bytes eine Blockgre von 4096.
Die Geschwindigkeit ist zumindest bei kleinen Blockgren vergleichbar mit AES-256 (CryptAES).


  Plain  ||  ||  ||  ||  ||    : ||    ||  ||  ||  ||  ||  Cipher
   ---->|XOR|->|ADD|->|Perm|->|XOR|->|ADD|---->|Perm|-->|XOR|->|ADD|->|Perm|->|XOR|->|ADD|---->
    8 |  |___|  |___|  |____|  |___|  |___| |  : |____| |  |___|  |___|  |____|  |___|  |___| | 8
      |    ^      ^      ^       ^      ^   |  :   ^    |    ^      ^      ^       ^      ^   |
      |    |7     |8     |2      |7     |8  |  :   |2   |    |7     |8     |2      |7     |8  |
      |    |      |      |       |      |   |  :   |    |    |      |      |       |      |   |
      |    |   ||   |   |  :   |    |    |   ||   |   |
    ||   ---|       MWC 1        |---  ||:   |  ||   ---|       MWC 2        |---  ||
    |i-1|----->| (256 Statusbits)   |<----|i-1|:    --|i+1|----->|  (256 Statusbits)  |<----|i+1|
    |___|   8  |____________________|  8  |___|:      |___|   8  |____________________|  8  |___|
                                               :
               Vorwrts,  i: 0..n-1            :                 Rckwrts,  i: n-1..0

 Perm: Permutation durch eine von vier Abbildungs-Tabellen, x = PTab[t][x], t: 0..3
 ADD:  8-Bit-Addition (Modulo 256)
 XOR:  8-Bit-Exklusiv-Oder
 i-1:  Zwischenspeicher/Verzgerungsglied vorwrts, bei i=0 Startwerte aus dem Schlssel
 i+1:  Zwischenspeicher/Verzgerungsglied rckwrts, bei i=n-1 Startwerte aus dem Schlssel
 MWC:  2 Pseudo-Zufallsgeneratoren, Periode je 2.6e76,  Multiply With Carry
       t=a*x+c+e, x=t%b, c=(int)(t/b), b=1<<32, e: 16-Bit-Einkopplung der Verzgerungsglieder
         a=A[k], x=X[k], c=C[k], k: 0..3 bei MWC 1 und 4..7 bei MWC 2, iterierend
*/

// Schlssel-Speicher, sollte nach Verwenden mit z.B. CryptBinBlock(0, 0, 0, "") gelscht werden
static BYTE byCBPermTab[4][256];
static BYTE byCBPermITab[4][256];
static DWORD pdwCBCryptSeed[16];
static DWORD dwLastInit;


int CryptBlock7KE(int iMode, BYTE* pData, int iLen)
{
	// verwendet byCBPermTab, byCBPermITab, tMWCStat, dwLastInit (Crypto-Key)
	// tMWCStat.x und tMWCStat.c wird dabei verndert und muss jeweils vor Aufruf neu initialisiert werden, z.B. mit
	//    memcpy(tMWCStat.x, pdwCBCryptSeed, 64);
	int ix;
	int iTab, iTab2, iOffs;
	DWORD dwLast = 0;
	DWORD dwMod = 0;
	BYTE* pDat;

	// Verschlsseln
	if (iMode & 1)
	{
		pDat = pData;
		dwLast = dwLastInit & 0xffff;
		for (ix=0; ix<iLen; ix++, pDat++)
		{
			dwMod = GetNextCryptModB(ix&3, dwLast);		// MWC-Generator, benutzt tMWCStat
			dwLast = *pDat;
			iTab = (dwMod>>30) & 0x03;
			iOffs = ((dwLast^(dwMod>>16 & 0x7f)) + (dwMod)) & 0xff;
			((BYTE*)&dwLast)[1] = *pDat = ((byCBPermTab[iTab][iOffs]^(BYTE)(dwMod>>23 & 0x7f)) + (BYTE)(dwMod>>8));
		}
		pDat = pData+iLen-1;
		dwLast = dwLastInit >> 16;
		for (ix=0; ix<iLen; ix++, pDat--)
		{
			dwMod = GetNextCryptModB(ix&3|4, dwLast);
			dwLast = byCBPermTab[dwLast&3][*pDat];	//++
			iTab = (dwMod>>30) & 0x03;
			iOffs = ((dwLast^(dwMod>>16 & 0x7f)) + (dwMod)) & 0xff;
			((BYTE*)&dwLast)[1] = *pDat = ((byCBPermTab[iTab][iOffs]^(BYTE)(dwMod>>23 & 0x7f)) + (BYTE)(dwMod>>8));
		}
	}

	// Entschlsseln
	else
	{
		pDat = pData+iLen-1;
		dwLast = dwLastInit >> 16;
		for (ix=0; ix<iLen; ix++, pDat--)
		{
			iTab2 = dwLast;
			dwMod = GetNextCryptModB(ix&3|4, dwLast);
			dwLast = *pDat;
			iTab = (dwMod>>30) & 0x03;
			iOffs = ((dwLast  - (BYTE)(dwMod>>8)) ^ (dwMod>>23 & 0x7f)) & 0xff;
			dwLast <<= 8;
			*(BYTE*)&dwLast = ((byCBPermITab[iTab][iOffs] - ((BYTE)dwMod))^(BYTE)(dwMod>>16 & 0x7f));
			*pDat = byCBPermITab[iTab2&3][*(BYTE*)&dwLast];	//++
		}
		pDat = pData;
		dwLast = dwLastInit & 0xffff;
		for (ix=0; ix<iLen; ix++, pDat++)
		{
			dwMod = GetNextCryptModB(ix&3, dwLast);
			dwLast = *pDat;
			iTab = (dwMod>>30) & 0x03;
			iOffs = ((dwLast  - (BYTE)(dwMod>>8)) ^ (dwMod>>23 & 0x7f)) & 0xff;
			dwLast <<= 8;
			((BYTE*)&dwLast)[0] = *pDat = ((byCBPermITab[iTab][iOffs] - ((BYTE)dwMod))^(BYTE)(dwMod>>16 & 0x7f));
		}
	}
	return 0;
}


int CryptBinBlock(int iMode, BYTE* pData, int iLen, char* szPassword, __int64 llBlockCnt, __int64 llBlockCntH, int iPasswordLen)
{
	int iTab;
	iKeyCntr = 0;	// Inistialisiert NextBinKey() beim nchsten Aufruf
	iKeyIx = 0;		// Inistialisiert GET_KEYBYTE(p) beim nchsten Aufruf
	dwDivKey = 0;	// Key-Divisor fr CreatePermutTab() initialisieren
	dwDivRef = 0;	// Referenz-Divisor fr CreatePermutTab() initialisieren

	if (pData==0 || iLen <= 0)
	{	// Schlsseldaten lschen
		memset(byCBPermTab, 0, sizeof(byCBPermTab));
		memset(byCBPermITab, 0, sizeof(byCBPermITab));
		memset(pdwCBCryptSeed, 0, sizeof(pdwCBCryptSeed));
		memset(tMWCStat.x, 0, 64);
		dwLastInit = 0;
		return 0;
	} 

	if (szPassword[0])	// Schlssel neu erzeugen
	{
		pszPasswordPtr = szPassword;	// Passwort fr SHA256-Key-Generator NextBinKey()
		if (iPasswordLen > 0) iPasswordPtrLen = iPasswordLen;
		else iPasswordPtrLen = strlen(szPassword);
		if (iMode & 0x20 && (llBlockCnt|llBlockCntH) != 0)
		{
			pszPasswordPtr = (PSTR)_alloca(iPasswordPtrLen + 18);
			strcpy(pszPasswordPtr, szPassword);
			*(__int64*)(pszPasswordPtr+iPasswordPtrLen) = llBlockCnt;
			iPasswordPtrLen += 8;
			if (llBlockCntH)
			{
				*(__int64*)(pszPasswordPtr+iPasswordPtrLen) = llBlockCntH;
				iPasswordPtrLen += 8;
			}
		}
		FillBinKey((PBYTE)pdwCBCryptSeed, 64);
		for (iTab=0; iTab<4; iTab++)
			CreatePermutTab(byCBPermTab[iTab], byCBPermITab[iTab], 256);

		dwDivKey <<= 8;
		GET_KEYBYTE(&dwDivKey);
		dwLastInit = dwDivKey;

		memset(bKeyBinary, 0, sizeof(bKeyBinary));
	}
	memcpy(tMWCStat.x, pdwCBCryptSeed, 64);

	if (!(iMode & 0x20) && (llBlockCnt != 0 || llBlockCntH != 0))
	{
		__int64 pllBuf[2] = { llBlockCnt, llBlockCntH };
		((__int64*)tMWCStat.x)[0] ^= pllBuf[1];	// sinnvoll, sonst ndert llBlockCntH den Vorwrts-MWC nicht oder nur teilweise
		((__int64*)tMWCStat.x)[1] ^= pllBuf[0];	// sinnvoll, sonst ndert llBlockCnt den Vorwrts-MWC nicht immer vollstndig
		CryptBlock7KE(1, (BYTE*)pllBuf, 16);	// modifiziert tMWCStat abhngig von BlockCnt und dem Key
	}

	CryptBlock7KE(iMode, pData, iLen);
	return 0;
}




