// CryptSha256.cpp

#include "stdafx.h"
#include <malloc.h>
#include <io.h>

#ifndef uint8
#define uint8  unsigned char
#endif

#ifndef uint32
#define uint32 unsigned long int
#endif

typedef struct
{
    uint32 total[2];
    uint32 state[8];
    uint8 buffer[64];
}
sha256_context;

void sha256_starts( sha256_context *ctx );
void sha256_update( sha256_context *ctx, uint8 *input, uint32 length );
void sha256_finish( sha256_context *ctx, uint8 digest[32] );


void Hash32Gen(PBYTE pData32, PBYTE pData=0, int iSize=0, PBYTE pHash32=0)
{
	sha256_context Sha_ctx;
	sha256_starts(&Sha_ctx);
	sha256_update(&Sha_ctx, pData32, 32);
	if (iSize) sha256_update(&Sha_ctx, pData, iSize);
	sha256_finish(&Sha_ctx, pHash32? pHash32:pData32);
}


BYTE ByteEncrypt(BYTE bData, PBYTE pKey, int iSize)
{
	BYTE b = bData;
	for (int i=0;; i++)
	{
		b = b + pKey[i] & 0x07 | (b & 0xf8) + (pKey[i] & 0xf8);
		if (i>=iSize-1) break;
		__asm rol byte ptr b,1
	}
	return b;
}

BYTE ByteDecrypt(BYTE bData, PBYTE pKey, int iSize)
{
	BYTE b = bData;
	for (int i=iSize-1;; i--)
	{
		b = b - pKey[i] & 0x07 | (b & 0xf8) - (pKey[i] & 0xf8);
		if (!i) break;
		__asm ror byte ptr b,1
	}
	return b;
}


int CryptSha256(int bEncrypt, PBYTE pDataIn, PBYTE pDataOut, DWORD dwSize, PBYTE pKey, int iKeySize, BYTE pInitValue[16])
{
    // Auf SHA256 basierende Verschsselung mit einer beliebigen Datengre
	// bEncrypt: 1=Verschlsseln, 0=Entschlsseln
	// pDataIn darf identisch mit pDataOut sein, beide haben die Gre dwSize
	// pKey zeigt auf den Schlssel der Gre iKeySize, blicherweie 32 Bytes
	// pInitValue ist entweder 0 oder ein Zeiger auf 16 Byte zum "salzen" der Daten,
	//   ist nicht geheim zu halten, sollt aber mglichst zufllig sein und
	//   muss beim Ver- und Entschlsseln gleich sein.
	//   Wird auch benutzt, um mehrere 16-Byte-Blcke zu verketten und ist dann
	//   jeweils der verschlsselte Vorgngerblock
	int i, k;
    unsigned char IV[32];
	unsigned char buffer[32];
    unsigned char SaveIn[16];
	sha256_context Sha_ctx;
	DWORD dwSizeRem = dwSize;

	cryptblock:
	if (dwSize > 16)	// Maximal 16 Bytes pro Aufruf
		dwSize = 16;
	if (dwSize == 0)
		return 0;

	sha256_starts(&Sha_ctx);
	if (pInitValue)
		sha256_update(&Sha_ctx, pInitValue, 16);
	sha256_update(&Sha_ctx, (PBYTE)&dwSizeRem, 4);
	sha256_update(&Sha_ctx, pKey, iKeySize);
	sha256_finish(&Sha_ctx, IV);

	if (dwSize == 1)
	{	// Spezialfall fr Gre 1
		if (bEncrypt)
			*pDataOut = ByteEncrypt(*pDataIn, IV, 32);
		else
			*pDataOut = ByteDecrypt(*pDataIn, IV, 32);
		return 0;
	}

	if (pDataOut != pDataIn)
		memcpy(pDataOut, pDataIn, dwSize);
	else if (!bEncrypt && dwSizeRem > 16)	// Sichern, falls noch bentigt
		memcpy(SaveIn, pDataIn, dwSize);
		
	int nh = (dwSize+1)>>1;
	int n = dwSize;
	int n2 = n - nh;
	int of = bEncrypt? 0x10 : 0;
	for (k=0; k<2; k++)
	{
		of ^= 0x10;
		if (bEncrypt)
		{
			Hash32Gen(IV, pDataOut, nh, buffer);
			for (i=nh; i<n; i++) pDataOut[i] += buffer[i];
			for (i=0;  i<n; i++) pDataOut[i] += IV[i+of];	// nur fr kommutative Operationen erlaubt
			Hash32Gen(IV, pDataOut+nh, n2, buffer);
			for (i=0; i<nh; i++) pDataOut[i] += buffer[i];
		}
		else
		{
			Hash32Gen(IV, pDataOut+nh, n2, buffer);
			for (i=0; i<nh; i++) pDataOut[i] -= buffer[i];
			for (i=0;  i<n; i++) pDataOut[i] -= IV[i+of];
			Hash32Gen(IV, pDataOut, nh, buffer);
			for (i=nh; i<n; i++) pDataOut[i] -= buffer[i];
		}
	}
	if (dwSizeRem > 16)
	{	// mehr als ein Block
		dwSizeRem -= 16;
		dwSize = dwSizeRem;
		memcpy(buffer, bEncrypt? pDataOut : pDataIn != pDataOut? pDataIn : SaveIn, 16);
		pInitValue = buffer;	// Verkettung mit altem Block
		pDataIn += 16;
		pDataOut += 16;
		goto cryptblock;
	}
	return 0;
}


